package Environment;

import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;

/**
 * Diese Klasse repräsentiert ein Environment durch das sich der Roboter bewegen
 * kann in Form eines Graphen. Ein Knoten wird hier als "Tile" repräsentert.
 * Jedes Tile kann mit bis zu 4 angrenzenden Tiles verbunden sein, welche sich
 * jeweils in einer der 4 Himmelsrichtungen befinden. Ein Environment-Graph wird
 * aus einer .txt File erzeugt.
 * 
 * Ein Environment wird grundsätzlich als rechteckig angenommen. Die Höhe des
 * Environment ergibt sich aus der Anzahl Zeilen des Textfiles, die Breite aus
 * der Länge der ersten Zeile. Ein Environment muss grundsätzlich vollständig
 * von BLOCKED Tiles umschlossen sein.
 * 
 * @author marco
 * 
 */
public class Environment
{
	// Zulässige chars für die verschiedenen Environment-Tiles im .txt-File
	private final char TILE_BLOCKED = 'x';
	private final char TILE_GOAL = 'g';
	private final char TILE_START = 's';
	private final char TILE_GROUND = ' ';
	private final char TILE_PATH = '\u00a4';

	private ArrayList<Tile> _environment;
	private Tile _startTile;
	private int _colLength;

	public Environment()
	{
		_environment = new ArrayList<Tile>();
		_startTile = null;
	}

	public Tile getStartTile()
	{
		return _startTile;
	}

	/**
	 * Erzeugt einen Environment-Graphen aus der übergebenen Datei. Zulässige
	 * Environtment-chars sind 'X'/'x' = Blocked, 'S'/'s' = Start, 'G'/'g' =
	 * Goal. Alle anderen Zeichen werden per Default als Ground angenommen Ein
	 * Environment wird grundsätzlich als rechteckig angenommen. Die Höhe des
	 * Environment ergibt sich aus der Anzahl Zeilen des Textfiles, die Breite
	 * aus der Länge der ersten Zeile. Ein Environment muss grundsätzlich
	 * vollständig von BLOCKED Tiles umschlossen sein.
	 * 
	 * @param path
	 *            der Pfad der Environment Textdatei
	 */
	public void buildEnvironmentFromFile(Path path)
	{
		assert path != null : "Path darf nicht null sein!";

		List<String> rows = null;
		Tile tile;
		Tile addedTile;

		// Die Textdatei zeilenweise einlesen und jede Zeile als StringListe
		// speichern
		try
		{
			rows = Files.readAllLines(path, Charset.defaultCharset());
		} catch (IOException ex)
		{
			ex.printStackTrace();
		}

		_colLength = rows.get(0).length();

		for (int row = 0; row < rows.size(); ++row)
		{
			for (int col = 0; col < _colLength; ++col)
			{
				char tyleChar = rows.get(row).toLowerCase().charAt(col);
				// Tiles vom jeweilige Typ (abhängig vom jeweiligen char)
				// erstellen und in die Environmentliste packen
				switch (tyleChar)
				{
					case TILE_START:
						_startTile = new Tile(row, col, TileType.START);
						_environment.add(_startTile);
						break;
					case TILE_GOAL:
						_environment.add(new Tile(row, col, TileType.GOAL));
						break;
					case TILE_BLOCKED:
						_environment.add(new Tile(row, col, TileType.BLOCKED));
						break;
					default:
						_environment.add(new Tile(row, col, TileType.GROUND));
				}
			}
		}

		// Für jedes Tile überprüfen, welche Tiles jeweils an dieses Tile
		// angrenzen und diese Tiles dann entsprechend miteinerander verbinden.
		// Ein Tile vom Typ BLOCKED befindet sich zwar in der Environment-Liste,
		// wird aber mit keinen weiteren Tiles verbunden, da es nicht erreichbar
		// ist
		for (int row = 0; row < rows.size(); ++row)
		{
			for (int col = 0; col < _colLength; ++col)
			{
				if ((row > 0)
						&& !(_environment.get(((row - 1) * _colLength) + col)
								.getType().equals(TileType.BLOCKED)))
				{
					tile = _environment.get((row * _colLength) + col);
					addedTile = _environment
							.get(((row - 1) * _colLength) + col);

					tile.addTile(addedTile, Direction.NORTH);
				}
				if ((row < rows.size() - 1)
						&& !(_environment.get(((row + 1) * _colLength) + col)
								.getType().equals(TileType.BLOCKED)))
				{
					tile = _environment.get((row * _colLength) + col);
					addedTile = _environment.get((row + 1) * _colLength + col);

					tile.addTile(addedTile, Direction.SOUTH);
				}
				if ((col > 0)
						&& !(_environment.get((row * _colLength) + col - 1)
								.getType().equals(TileType.BLOCKED)))
				{
					tile = _environment.get((row * _colLength) + col);
					addedTile = _environment.get((row * _colLength) + col - 1);

					tile.addTile(addedTile, Direction.WEST);
				}
				if ((col < (rows.get(0).length() - 1))
						&& !(_environment.get((row * _colLength) + col + 1)
								.getType().equals(TileType.BLOCKED)))
				{
					tile = _environment.get((row * _colLength) + col);
					addedTile = _environment.get((row * _colLength) + col + 1);

					tile.addTile(addedTile, Direction.EAST);
				}
			}
		}
	}

	/**
	 * Gibt das Environment mitsamt dem gefundenen Pfad zum Ziel auf der Konsole
	 * aus
	 * 
	 * @param foundPath
	 *            der gefundene Pfad vom Start zum Ziel
	 */
	public void toConsole(ArrayList<Tile> foundPath)
	{
		for (int i = 0; i < _environment.size(); ++i)
		{
			if ((i % _colLength) == 0)
			{
				System.out.println();
			}

			switch (_environment.get(i).getType())
			{
				case START:
					System.out.print(TILE_START);
					break;
				case GOAL:
					System.out.print(TILE_GOAL);
					break;
				case BLOCKED:
					System.out.print(TILE_BLOCKED);
					break;
				case PATH:
					System.out.print(TILE_PATH);
					break;
				default:
					System.out.print(TILE_GROUND);
					break;
			}

		}

	}

}
